import datetime
import unittest
from unittest.mock import patch

import requests
from pydantic import ValidationError

from weave.trace_server import trace_server_interface as tsi
from weave.trace_server.ids import generate_id
from weave.trace_server_bindings.remote_http_trace_server import RemoteHTTPTraceServer


def generate_start(id) -> tsi.StartedCallSchemaForInsert:
    return tsi.StartedCallSchemaForInsert(
        project_id="test",
        id=id or generate_id(),
        op_name="test_name",
        trace_id="test_trace_id",
        parent_id="test_parent_id",
        started_at=datetime.datetime.now(tz=datetime.timezone.utc)
        - datetime.timedelta(seconds=1),
        attributes={"a": 5},
        inputs={"b": 5},
    )


class TestRemoteHTTPTraceServer(unittest.TestCase):
    def setUp(self):
        self.trace_server_url = "http://example.com"
        self.server = RemoteHTTPTraceServer(self.trace_server_url)

    @patch("weave.trace_server.requests.post")
    def test_ok(self, mock_post):
        call_id = generate_id()
        mock_post.return_value = requests.Response()
        mock_post.return_value.json = lambda: dict(
            tsi.CallStartRes(id=call_id, trace_id="test_trace_id")
        )
        mock_post.return_value.status_code = 200
        start = generate_start(call_id)
        self.server.call_start(tsi.CallStartReq(start=start))
        mock_post.assert_called_once()

    @patch("weave.trace_server.requests.post")
    def test_400_500_no_retry(self, mock_post):
        call_id = generate_id()
        resp1 = requests.Response()
        resp1.json = lambda: dict(
            tsi.CallStartRes(id=call_id, trace_id="test_trace_id")
        )
        resp1.status_code = 400

        resp2 = requests.Response()
        resp2.status_code = 500

        mock_post.side_effect = [
            resp1,
            resp2,
        ]

        start = generate_start(call_id)
        with self.assertRaises(requests.HTTPError):
            self.server.call_start(tsi.CallStartReq(start=start))

        with self.assertRaises(requests.HTTPError):
            self.server.call_start(tsi.CallStartReq(start=start))

    def test_invalid_no_retry(self):
        with self.assertRaises(ValidationError):
            self.server.call_start(tsi.CallStartReq(start={"invalid": "broken"}))

    @patch("weave.trace_server.requests.post")
    def test_502_503_504_429_retry(self, mock_post):
        call_id = generate_id()

        resp1 = requests.Response()
        resp1.status_code = 502

        resp2 = requests.Response()
        resp2.status_code = 503

        resp3 = requests.Response()
        resp3.status_code = 504

        resp4 = requests.Response()
        resp4.status_code = 429

        resp5 = requests.Response()
        resp5.json = lambda: dict(
            tsi.CallStartRes(id=call_id, trace_id="test_trace_id")
        )
        resp5.status_code = 200

        mock_post.side_effect = [resp1, resp2, resp3, resp4, resp5]
        start = generate_start(call_id)
        self.server.call_start(tsi.CallStartReq(start=start))

    @patch("weave.trace_server.requests.post")
    def test_other_error_retry(self, mock_post):
        call_id = generate_id()

        resp2 = requests.Response()
        resp2.json = lambda: dict(
            tsi.CallStartRes(id=call_id, trace_id="test_trace_id")
        )
        resp2.status_code = 200

        mock_post.side_effect = [
            ConnectionResetError(),
            ConnectionError(),
            OSError(),
            TimeoutError(),
            resp2,
        ]
        start = generate_start(call_id)
        self.server.call_start(tsi.CallStartReq(start=start))


if __name__ == "__main__":
    unittest.main()



import functools
import typing
from typing import Callable, Optional, TypeVar

from typing_extensions import ParamSpec

from weave.legacy.weave import weave_types as types
from weave.legacy.weave import context_state, derive_op, op_args, op_def, pyfunc_type_util, registry_mem

if typing.TYPE_CHECKING:
    from weave.legacy.weave.gql_op_plugin import GqlOpPlugin


# Important usability note
#
# Even though the op decorator actually returns an OpDef object, the
# type annotation of the return is just Callable. If the signature
# is changed back to OpDef, pyright (VSCode's language server) will
# not pass through the underlying function type signature. So for
# example, hovering an op decorated function will never show the original
# function's docstring. At runtime, we call functools.update_wrapper
# to copy the original function's signature to the OpDef object, but
# the language server's static analysis does not see this.
#
# I tried to use a Protocol that extends Callable to mix OpDef attributes
# in with Callable, but that also makes the decorator opaque to the
# language server.
#
# So we make a specifity tradeoff here: decorated functions will look
# just like their original functions to type checkers, and users will
# need to cast them to OpDef if they want to access the OpDef attributes
# in a typesafe way. You can use the weave.as_op_def() helper to do this.

P = ParamSpec("P")
R = TypeVar("R")


def op(
    input_type: pyfunc_type_util.MaybeInputTypeType = None,
    # input_type: pyfunc_type_util.MaybeInputTypeType = None,
    output_type: pyfunc_type_util.MaybeOutputTypeType = None,
    refine_output_type: Optional[op_def.OpDef] = None,
    name: Optional[str] = None,
    setter: Optional[Callable] = None,
    render_info: Optional[dict] = None,
    hidden: bool = False,
    pure: bool = True,
    _op_def_class: type[op_def.OpDef] = op_def.OpDef,
    *,  # Marks the rest of the arguments as keyword-only.
    plugins: Optional[dict[str, "GqlOpPlugin"]] = None,
    mutation: bool = False,
    # If True, the op will be weavified, ie it's resolver will be stored as a Weave
    # op graph. The compile node_ops pass will expand the node to the weavified
    # version, instead of executing the original python resolver body.
    weavify: bool = False,
) -> Callable[[Callable[P, R]], Callable[P, R]]:
    """Decorator for declaring an op.

    Decorated functions must be typed, either with Python types or by declaring
    input_type, output_type as arguments to op (Python types preferred).
    """

    # For builtins, enforce that parameter and return types must be declared.
    # For user ops, allow missing types.
    loading_builtins = context_state._loading_built_ins.get()
    allow_unknowns = not loading_builtins

    def wrap(f: Callable[P, R]) -> Callable[P, R]:
        weave_input_type = pyfunc_type_util.determine_input_type(
            f, input_type, allow_unknowns=allow_unknowns
        )
        weave_output_type = pyfunc_type_util.determine_output_type(
            f, output_type, allow_unknowns=allow_unknowns
        )

        fq_op_name = name
        if fq_op_name is None:
            op_prefix = "op"

            # This would be a much nicer: automatically use first type name.
            # But what about Unions? I'm not going to solve in this PR.
            # TODO: implement this.
            # if (
            #     isinstance(weave_input_type, op_args.OpNamedArgs)
            #     and weave_input_type.arg_types
            # ):
            #     arg_type0 = next(iter(weave_input_type.arg_types.values()))
            #     if arg_type0 != types.UnknownType():
            #         op_prefix = arg_type0.name

            fq_op_name = f"{op_prefix}-{f.__name__}"

        op = _op_def_class(
            fq_op_name,
            weave_input_type,
            weave_output_type,
            f,
            refine_output_type=refine_output_type,
            setter=setter,
            render_info=render_info,
            hidden=hidden,
            pure=pure,
            plugins=plugins,
            mutation=mutation,
        )
        if weavify:
            from weave.legacy.weave.weavify import op_to_weave_fn

            op.weave_fn = op_to_weave_fn(op)

        # We shouldn't bother registering for weaveflow. We could register
        # only in the case where an op is lazily called. But no need for now.
        op_version = registry_mem.memory_registry.register_op(op)

        if loading_builtins:
            derive_op.derive_ops(op)

        functools.update_wrapper(op_version, f)

        return op_version

    return wrap



"""
This file contains the data structures and methods used to manage the in-memory
state of tagged objects. Crticailly, it relies on two private global objects:

* `_OBJ_TAGS_MEM_MAP` - used to map the python id of an object to a dictionary of
    tags. This is used to store the tags for an object.
* `_VISITED_OBJ_IDS` - used to keep track of the python ids of objects that are
    currently being visited. This is used to prevent infinite recursion when
    determing the type of tagged objects.

The primary user-facing functions are:

* `add_tags` - used to add tags to an object
* `find_tag` - used to recursively lookup the tag for a given object

"""

import contextvars
import logging
import typing
import weakref
from collections import defaultdict
from contextlib import contextmanager

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave import engine_trace, box

statsd = engine_trace.statsd()  # type: ignore

NodeTagStoreType = dict[int, dict[str, typing.Any]]
TagStoreType = defaultdict[int, NodeTagStoreType]


# Private global objects used to store the tags for objects
_OBJ_TAGS_MEM_MAP: contextvars.ContextVar[
    typing.Optional[TagStoreType]  # shape: {node_id: {obj_id: {tag_key: tag_value}}}
] = contextvars.ContextVar("obj_tags_mem_map", default=None)

# Current node id for scoping tags
_OBJ_TAGS_CURR_NODE_ID: contextvars.ContextVar[int] = contextvars.ContextVar(
    "obj_tags_curr_node", default=-1
)


# gets the current tag memory map for the current node
def _current_obj_tag_mem_map() -> typing.Optional[NodeTagStoreType]:
    node_tags = _OBJ_TAGS_MEM_MAP.get()
    if node_tags is None:
        return None
    return node_tags[_OBJ_TAGS_CURR_NODE_ID.get()]


def current_tag_store_size() -> int:
    current_mmap = _OBJ_TAGS_MEM_MAP.get()
    if current_mmap is not None:
        n_tag_store_entries = sum(len(current_mmap[key]) for key in current_mmap)
    else:
        n_tag_store_entries = 0
    return n_tag_store_entries


def record_current_tag_store_size() -> None:
    n_tag_store_entries = current_tag_store_size()
    logging.info(f"Current number of tag store entries: {n_tag_store_entries}")
    statsd.gauge("weave.tag_store.num_entries", n_tag_store_entries)


@contextmanager
def with_tag_store_state(
    curr_node_id: int, tags_mem_map: typing.Optional[TagStoreType]
) -> typing.Iterator[None]:
    tag_store_token = _OBJ_TAGS_MEM_MAP.set(tags_mem_map)
    curr_node_id_token = _OBJ_TAGS_CURR_NODE_ID.set(curr_node_id)
    yield
    _OBJ_TAGS_CURR_NODE_ID.reset(curr_node_id_token)
    _OBJ_TAGS_MEM_MAP.reset(tag_store_token)


# sets the current node with optionally merged in parent tags
@contextmanager
def set_curr_node(node_id: int, parent_node_ids: list[int]) -> typing.Iterator[None]:
    node_tags = _OBJ_TAGS_MEM_MAP.get()
    if node_tags is None:
        raise errors.WeaveInternalError("No tag store context")
    token = _OBJ_TAGS_CURR_NODE_ID.set(node_id)
    for parent_id in parent_node_ids:
        node_tags[node_id].update(node_tags[parent_id])
    try:
        yield None
    finally:
        _OBJ_TAGS_CURR_NODE_ID.reset(token)


# Private global objects used to keep track of the python ids of objects that are
# currently being visited.
_VISITED_OBJ_IDS: contextvars.ContextVar[set[int]] = contextvars.ContextVar(
    "visited_obj_ids", default=set()
)


# Callers can create an isolated tagging context by using this context manager
# This is primarily used by the executor to prevent tags from leaking between
# different executions. See execute.py for it's usage.
# Only creates a new context for the top-level call in the stack. Re-entrant calls
# will use the same context.
@contextmanager
def isolated_tagging_context() -> typing.Iterator[None]:
    created_context = False
    if _OBJ_TAGS_MEM_MAP.get() is None:
        created_context = True
        token = _OBJ_TAGS_MEM_MAP.set(defaultdict(dict))
    try:
        yield None
    finally:
        if created_context:
            _OBJ_TAGS_MEM_MAP.reset(token)


@contextmanager
def new_tagging_context() -> typing.Iterator[None]:
    token = _OBJ_TAGS_MEM_MAP.set(defaultdict(dict))
    try:
        yield None
    finally:
        _OBJ_TAGS_MEM_MAP.reset(token)


# Callers can indicate that an object is being visited by using this context manager
# This is primarily used by the TaggedValueType::type_of_instance method to prevent
# infinite recursion when determining the type of tagged objects.
@contextmanager
def with_visited_obj(obj: typing.Any) -> typing.Iterator[None]:
    id_val = get_id(obj)
    assert id_val == get_id(box.box(obj)), "Can only tag boxed objects"
    visited_obj_ids = _VISITED_OBJ_IDS.get()
    visited_obj_ids.add(id_val)
    try:
        yield None
    finally:
        visited_obj_ids.remove(id_val)


def _remove_tags(mem_map: dict[int, typing.Any], id_val: int) -> None:
    node_tags = _OBJ_TAGS_MEM_MAP.get()
    if node_tags is None:
        return
    # The tags can be on any node, so we need to check all of them.
    # This is algorithmically inefficient and could be a serious performance
    # problem.
    for node_id, tag_map in node_tags.items():
        if id_val in tag_map:
            del tag_map[id_val]


def get_id(obj: typing.Any) -> int:
    if box.cannot_have_weakref(obj):
        if obj._id is not None:
            return obj._id
        obj._id = box.make_id()
        return obj._id
    return id(obj)


# Adds a dictionary of tags to an object
def add_tags(
    obj: typing.Any,
    tags: dict[str, typing.Any],
    give_precedence_to_existing_tags: bool = False,
) -> typing.Any:
    mem_map = _current_obj_tag_mem_map()
    if mem_map is None:
        raise errors.WeaveInternalError("No tag store context")
    id_val = get_id(obj)
    if not box.cannot_have_weakref(obj) and id_val not in mem_map:
        # Ensure we cleanup the tags when the object is garbage collected.
        # Python is happy to reuse IDs after they are freed!
        try:
            weakref.finalize(obj, _remove_tags, mem_map, id_val)
        except:
            # Extreme bug here!
            # Can't box pydantic objects, like those from langchain.
            # TODO: fix
            pass
    assert box.is_boxed(obj), "Can only tag boxed objects"
    existing_tags = get_tags(obj) if is_tagged(obj) else {}
    if give_precedence_to_existing_tags:
        mem_map[id_val] = {**tags, **existing_tags}
    else:
        mem_map[id_val] = {**existing_tags, **tags}
    return obj


# Gets the dictionary of tags assocaited with the given object
# Note: this is not recursive, it only returns the tags directly assocaited with
# the given object
def get_tags(obj: typing.Any) -> dict[str, typing.Any]:
    id_val = get_id(obj)
    if id_val in _VISITED_OBJ_IDS.get():
        raise ValueError("Cannot get tags for an object that is being visited")

    current_mem_map = _current_obj_tag_mem_map()
    if current_mem_map is None:
        return {}
    if id_val not in current_mem_map:
        return {}
    return current_mem_map[id_val]


# Recursively looks up the tag for the object, given a key and target tag_type.
def find_tag(
    obj: typing.Any, key: str, tag_type: types.Type = types.Any()
) -> typing.Any:
    # TODO: Implement tag type filtering using tag_type
    cur_tags = get_tags(obj)
    if key in cur_tags:
        return cur_tags[key]
    else:
        for parent in cur_tags.values():
            if is_tagged(parent):
                par_tag = find_tag(parent, key)
                if par_tag is not None:
                    return par_tag
    return None


# Returns true if the given object has been tagged
def is_tagged(obj: typing.Any) -> bool:
    id_val = get_id(obj)
    if id_val in _VISITED_OBJ_IDS.get():
        return False
    mem_map = _current_obj_tag_mem_map()
    if mem_map is None:
        return False

    return id_val in mem_map


def clear_tag_store() -> None:
    tag_store = _OBJ_TAGS_MEM_MAP.get()
    cur_obj_tag_mem_map = _OBJ_TAGS_CURR_NODE_ID.get()

    if tag_store is not None:
        tag_store.clear()

    if cur_obj_tag_mem_map is not None:
        _OBJ_TAGS_CURR_NODE_ID.set(-1)



import cProfile
import json
import re

import numpy as np
import pytest
import wandb

from weave.legacy.weave import api as weave
from weave.legacy.weave import artifact_fs, artifact_wandb, compile, graph, ops, stitch, uris
from weave.legacy.weave import ops_arrow as arrow
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.language_features.tagging.tagged_value_type import TaggedValueType
from weave.legacy.weave.ops_arrow import ArrowWeaveListType
from weave.legacy.weave.ops_domain import artifact_membership_ops as amo
from weave.legacy.weave.ops_domain import table, wb_util, wbmedia
from weave.legacy.weave.ops_domain import wb_domain_types as wdt
from weave.legacy.weave.ops_primitives import dict_, list_
from weave.legacy.weave.ops_primitives.file import _as_w0_dict_
from .test_wb_domain_ops import assert_gql_str_equal

from ...tests import fixture_fakewandb as fwb
from weave.legacy.tests.util import weavejs_ops

file_path_response = {
    "project_518fa79465d8ffaeb91015dce87e092f": {
        **fwb.project_payload,  # type: ignore
        "artifactType_46d22fef09db004187bb8da4b5e98c58": {
            **fwb.defaultArtifactType_payload,  # type: ignore
            "artifactCollections_c1233b7003317090ab5e2a75db4ad965": {
                "edges": [
                    {
                        "node": {
                            **fwb.artifactSequence_payload,  # type: ignore
                            "artifacts_c1233b7003317090ab5e2a75db4ad965": {
                                "edges": [{"node": fwb.artifactVersion_payload}]
                            },
                        }
                    }
                ]
            },
        },
    }
}

file_path_no_entity_response = {
    "project_518fa79465d8ffaeb91015dce87e092f": {
        **fwb.project_payload,  # type: ignore
        "artifactType_46d22fef09db004187bb8da4b5e98c58": {
            **fwb.defaultArtifactType_payload,  # type: ignore
            "artifactCollections_c1233b7003317090ab5e2a75db4ad965": {
                "edges": [
                    {
                        "node": {
                            **fwb.artifactSequence_payload,  # type: ignore
                            "artifacts_c1233b7003317090ab5e2a75db4ad965": {
                                "edges": [
                                    {"node": fwb.artifactVersion_no_entity_payload}
                                ]
                            },
                        }
                    }
                ]
            },
        },
    }
}

artifact_browser_response = {
    "project_518fa79465d8ffaeb91015dce87e092f": {
        **fwb.project_payload,  # type: ignore
        "artifactCollection_d651817074b6a8074e87e9dfd5767726": {
            **fwb.artifactSequence_payload,  # type: ignore
            "artifactMembership_78e7a3fd51159b4fdfb0815be0b0f92c": fwb.artifactMembership_payload,
        },
    }
}

workspace_response_run1_summary_metrics = json.dumps(
    {
        "table": {
            "_type": "table-file",
            "artifact_path": "wandb-client-artifact://1234567890/test_results.table.json",
        },
        "legacy_table": {
            "_type": "table-file",
            "path": "media/tables/legacy_table.table.json",
        },
        "image": {
            "height": 64,
            "sha256": "440fab0d6f537b4557a106fa7853453332650631ef580fd328c620bd8aa5a025",
            "path": "media/images/random_image_9_440fab0d6f537b4557a1.png",
            "size": 4228,
            "_type": "image-file",
            "width": 64,
            "format": "png",
        },
    }
)


workspace_response_run3_summary_metrics = json.dumps(
    {
        "table": {
            "_type": "table-file",
            "artifact_path": "wandb-client-artifact://1122334455/test_results.table.json",
        },
    }
)


def workspace_response(include_display_name=True):
    return {
        "project_518fa79465d8ffaeb91015dce87e092f": {
            **fwb.project_payload,  # type: ignore
            "runs_c1233b7003317090ab5e2a75db4ad965": {
                "edges": [
                    {
                        "node": {
                            **fwb.run_payload,  # type: ignore
                            "summaryMetricsSubset": workspace_response_run1_summary_metrics,
                            "summaryMetrics": workspace_response_run1_summary_metrics,
                            **(
                                {"displayName": "amber-glade-100"}
                                if include_display_name
                                else {}
                            ),
                        },
                    },
                    {
                        "node": {
                            **fwb.run2_payload,  # type: ignore
                            "summaryMetricsSubset": json.dumps({}),
                            "summaryMetrics": json.dumps({}),
                            **(
                                {"displayName": "run2-display_name"}
                                if include_display_name
                                else {}
                            ),
                        },
                    },
                    {
                        "node": {
                            **fwb.run3_payload,  # type: ignore
                            "summaryMetricsSubset": workspace_response_run3_summary_metrics,
                            "summaryMetrics": workspace_response_run3_summary_metrics,
                            **(
                                {"displayName": "run3-display_name"}
                                if include_display_name
                                else {}
                            ),
                        },
                    },
                ]
            },
        }
    }


workspace_response_no_run_displayname_run1_summary_metrics = json.dumps(
    {
        "table": {
            "_type": "table-file",
            "artifact_path": "wandb-client-artifact://1234567890/test_results.table.json",
        },
        "legacy_table": {
            "_type": "table-file",
            "path": "media/tables/legacy_table.table.json",
        },
    }
)

workspace_response_no_run_displayname_run3_summary_metrics = json.dumps(
    {
        "table": {
            "_type": "table-file",
            "artifact_path": "wandb-client-artifact://1122334455/test_results.table.json",
        },
    }
)

workspace_response_no_run_displayname = {
    "project_518fa79465d8ffaeb91015dce87e092f": {
        **fwb.project_payload,  # type: ignore
        "runs_c1233b7003317090ab5e2a75db4ad965": {
            "edges": [
                {
                    "node": {
                        **fwb.run_payload,  # type: ignore
                        "summaryMetricsSubset": workspace_response_no_run_displayname_run1_summary_metrics,
                        "summaryMetrics": workspace_response_no_run_displayname_run1_summary_metrics,
                    },
                },
                {
                    "node": {
                        **fwb.run2_payload,  # type: ignore
                        "summaryMetricsSubset": json.dumps({}),
                        "summaryMetrics": json.dumps({}),
                    },
                },
                {
                    "node": {
                        **fwb.run3_payload,  # type: ignore
                        "summaryMetricsSubset": workspace_response_no_run_displayname_run3_summary_metrics,
                        "summaryMetrics": workspace_response_no_run_displayname_run3_summary_metrics,
                    },
                },
            ]
        },
    }
}


empty_workspace_response = {
    "project_518fa79465d8ffaeb91015dce87e092f": {
        **fwb.project_payload,  # type: ignore
        "runs_c1233b7003317090ab5e2a75db4ad965": {"edges": []},
    }
}

workspace_response_filtered_run1_summary_metrics = json.dumps(
    {
        "table": {
            "_type": "table-file",
            "artifact_path": "wandb-client-artifact://1234567890/test_results.table.json",
        }
    }
)

workspace_response_filtered = {
    "project_518fa79465d8ffaeb91015dce87e092f": {
        **fwb.project_payload,  # type: ignore
        "runs_261949318143369aa6c158af92afee03": {
            "edges": [
                {
                    "node": {
                        **fwb.run_payload,  # type: ignore
                        "summaryMetricsSubset": workspace_response_filtered_run1_summary_metrics,
                        "summaryMetrics": workspace_response_filtered_run1_summary_metrics,
                        "displayName": "amber-glade-100",
                    }
                },
            ]
        },
    }
}

project_run_artifact_response = {
    "project_518fa79465d8ffaeb91015dce87e092f": {
        **fwb.project_payload,  # type: ignore
        "runs_c1233b7003317090ab5e2a75db4ad965": {
            "edges": [
                {
                    "node": {
                        **fwb.run_payload,  # type: ignore
                        "outputArtifacts_c1233b7003317090ab5e2a75db4ad965": {
                            "edges": [
                                {
                                    "node": {
                                        "id": "QXJ0aWZhY3Q6MTQyNTg4OTM=",
                                        "versionIndex": 0,
                                        "artifactSequence": fwb.artifactSequence_payload,
                                    }
                                }
                            ]
                        },
                    }
                }
            ]
        },
    }
}

artifact_version_sdk_response = {
    "artifact": {
        **fwb.artifactVersion_payload,  # type: ignore
        "state": "COMMITTED",
        "commitHash": "303db33c9f9264768626",
        "artifactType": fwb.defaultArtifactType_payload,
        "artifactSequence": {
            **fwb.artifactSequence_payload,
            "project": fwb.project_payload,
            "state": "READY",
        },  # type: ignore
    }
}


@pytest.mark.parametrize(
    "table_file_node_fn, mock_response",
    [
        # Path used in weave demos
        (
            (
                lambda: ops.project("stacey", "mendeleev")
                .artifactType("test_results")
                .artifacts()[0]
                .versions()[0]
                .file("test_results.table.json")
            ),
            file_path_response,
        ),
        # Path used in artifact browser
        (
            (
                lambda: ops.project("stacey", "mendeleev")
                .artifact("test_res_1fwmcd3q")
                .membershipForAlias("v0")
                .artifactVersion()
                .file("test_results.table.json")
            ),
            artifact_browser_response,
        ),
    ],
)
def test_table_call(table_file_node_fn, mock_response, fake_wandb):
    fake_wandb.fake_api.add_mock(lambda q, ndx: mock_response)
    table_file_node = table_file_node_fn()
    table_image0_node = table_file_node.table().rows()[0]["image"]
    table_image0 = weave.use(table_image0_node)
    assert table_image0.height == 299
    assert table_image0.width == 299
    assert table_image0.path == "media/images/6274b7484d7ed4b6ad1b.png"


def test_avfile_type(fake_wandb):
    fake_wandb.fake_api.add_mock(lambda q, ndx: file_path_response)
    f = (
        ops.project("stacey", "mendeleev")
        .artifactType("test_results")
        .artifacts()[0]
        .versions()[0]
        .file("test_results.table.json")
    )
    t = weavejs_ops.file_type(f)
    assert TaggedValueType(
        types.TypedDict(property_types={"project": wdt.ProjectType}),
        artifact_fs.FilesystemArtifactFileType(
            extension=types.Const(types.String(), "json"),
            wbObjectType=ops.TableType(),
        ),
    ).assign_type(weave.use(t))


def test_table_col_order_and_unknown_types(fake_wandb):
    fake_wandb.fake_api.add_mock(lambda q, ndx: file_path_response)
    node = (
        ops.project("stacey", "mendeleev")
        .artifactType("test_results")
        .artifacts()[0]
        .versions()[0]
        .file("weird_table.table.json")
        .table()
    )
    assert weave.use(node.rows()[0]["c"]) == 9.93


def test_missing_file(fake_wandb):
    fake_wandb.fake_api.add_mock(lambda q, ndx: file_path_response)
    node = (
        ops.project("stacey", "mendeleev")
        .artifactType("test_results")
        .artifacts()[0]
        .versions()[0]
        .file("does_not_exist")
    )
    assert weave.use(node) == None


def test_missing_file_no_entity(fake_wandb):
    fake_wandb.fake_api.add_mock(lambda q, ndx: file_path_no_entity_response)
    node = (
        ops.project("stacey", "mendeleev")
        .artifactType("test_results")
        .artifacts()[0]
        .versions()[0]
        .file("does_not_exist")
    )
    assert weave.use(node) == None


def table_mock1(q, ndx):
    if q["gql"].definitions[0].name.value == "WeavePythonCG":
        return workspace_response()
    else:
        return artifact_version_sdk_response


def table_mock1_no_display_name(q, ndx):
    if q["gql"].definitions[0].name.value == "WeavePythonCG":
        return workspace_response(False)
    else:
        return artifact_version_sdk_response


def table_mock2(q, ndx):
    if q["gql"].definitions[0].name.value == "WeavePythonCG":
        return workspace_response()
    else:
        return artifact_version_sdk_response


def table_mock_empty_workspace(q, ndx):
    if q["gql"].definitions[0].name.value == "WeavePythonCG":
        return empty_workspace_response


def test_map_gql_op(fake_wandb):
    fake_wandb.fake_api.add_mock(lambda q, ndx: project_run_artifact_response)
    node = (
        ops.project("stacey", "mendeleev")
        .runs()
        .limit(1)
        .loggedArtifactVersions()
        .limit(1)[0][0]
        .name()
    )
    assert weave.use(node) == "test_res_1fwmcd3q:v0"


def test_legacy_run_file_table_format(fake_wandb, cache_mode_minimal):
    fake_wandb.fake_api.add_mock(table_mock1)
    cell_node = (
        ops.project("stacey", "mendeleev")
        .runs()
        .limit(1)
        .summary()["legacy_table"]
        .table()
        .rows()
        .dropna()
        .concat()
        .createIndexCheckpointTag()[1]["col1"]
    )
    assert weave.use(cell_node) == "c"
    assert weave.use(cell_node.indexCheckpoint()) == 1
    assert weave.use(cell_node.run().name()) == "amber-glade-100"
    assert weave.use(cell_node.project().name()) == "mendeleev"


def test_mapped_table_empty(fake_wandb, cache_mode_minimal):
    fake_wandb.fake_api.add_mock(table_mock_empty_workspace)
    cell_node = (
        ops.project("stacey", "mendeleev")
        .runs()
        .limit(1)
        .summary()["table"]
        .table()
        .rows()
        .dropna()
        .concat()
        .createIndexCheckpointTag()[5]["score_Amphibia"]
    )
    assert weave.use(cell_node.indexCheckpoint()) == None
    assert weave.use(cell_node.run().name()) == None
    assert weave.use(cell_node.project().name()) == "mendeleev"


def test_mapped_table_tags(fake_wandb, cache_mode_minimal):
    fake_wandb.fake_api.add_mock(table_mock1)
    cell_node = (
        ops.project("stacey", "mendeleev")
        .runs()
        .limit(1)
        .summary()["table"]
        .table()
        .rows()
        .dropna()
        .concat()
        .createIndexCheckpointTag()[5]["score_Amphibia"]
    )
    assert weave.use(cell_node.indexCheckpoint()) == 5
    assert weave.use(cell_node.run().name()) == "amber-glade-100"
    assert weave.use(cell_node.project().name()) == "mendeleev"


def test_table_tags_row_first(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock1)
    cell_node = (
        ops.project("stacey", "mendeleev")
        .runs()
        .limit(10)[0]
        .summary()["table"]
        .table()
        .rows()
        .createIndexCheckpointTag()[5]["score_Amphibia"]
    )
    assert weave.use(cell_node.indexCheckpoint()) == 5
    assert weave.use(cell_node.run().name()) == "amber-glade-100"
    assert weave.use(cell_node.project().name()) == "mendeleev"


def test_workspace_table_type(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock2)
    summary_node = ops.project("stacey", "mendeleev").runs().summary()
    cell_node = summary_node["table"].table()
    assert TaggedValueType(
        types.TypedDict(property_types={"project": wdt.ProjectType}),
        types.List(
            object_type=types.optional(
                TaggedValueType(
                    types.TypedDict(property_types={"run": wdt.RunType}),
                    types.optional(ops.TableType()),
                )
            )
        ),
    ).assign_type(cell_node.type)


def test_workspace_table_rows_type(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock2)
    cell_node = (
        ops.project("stacey", "mendeleev").runs().summary()["table"].table().rows()
    )
    assert TaggedValueType(
        types.TypedDict(property_types={"project": wdt.ProjectType}),
        types.List(
            object_type=types.optional(
                TaggedValueType(
                    types.TypedDict(property_types={"run": wdt.RunType}),
                    ops.ArrowWeaveListType(
                        object_type=types.TypedDict(
                            property_types={
                                "id": types.optional(types.String()),
                                "image": types.optional(
                                    ops.ImageArtifactFileRef.WeaveType()
                                ),
                                "guess": types.optional(types.String()),
                                "truth": types.optional(types.String()),
                                "score_Animalia": types.optional(types.Number()),
                                "score_Amphibia": types.optional(types.Number()),
                                "score_Arachnida": types.optional(types.Number()),
                                "score_Aves": types.optional(types.Number()),
                                "score_Fungi": types.optional(types.Number()),
                                "score_Insecta": types.optional(types.Number()),
                                "score_Mammalia": types.optional(types.Number()),
                                "score_Mollusca": types.optional(types.Number()),
                                "score_Plantae": types.optional(types.Number()),
                                "score_Reptilia": types.optional(types.Number()),
                            }
                        )
                    ),
                )
            )
        ),
    ).assign_type(cell_node.type)


def test_table_tags_column_first(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock1)
    cell_node = (
        ops.project("stacey", "mendeleev")
        .runs()
        .limit(1)[0]
        .summary()["table"]
        .table()
        .rows()
        .createIndexCheckpointTag()["score_Amphibia"][5]
    )
    assert weave.use(cell_node.indexCheckpoint()) == 5
    assert weave.use(cell_node.run().name()) == "amber-glade-100"
    assert weave.use(cell_node.project().name()) == "mendeleev"


def test_table_images(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock2)
    # Query 1:
    project_node = ops.project("stacey", "mendeleev")
    project_runs_node = project_node.runs().limit(1)
    summary_node = project_runs_node.summary()
    # this use is important as it models the sequence of calls in UI
    # and invokes the issues with artifact paths
    assert list(weave.use(summary_node)[0].keys()) == ["table", "legacy_table", "image"]

    # Query 2:
    table_rows_node = summary_node.pick("table").table().rows()
    assert len(weave.use(table_rows_node)) == 1


def table_mock_filtered(q, ndx):
    if q["gql"].definitions[0].name.value == "WeavePythonCG":
        return workspace_response_filtered
    else:
        return artifact_version_sdk_response


def test_tag_run_color_lookup(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock_filtered)
    colors_node = weave.save(
        {"1ht5692d": "rgb(218, 76, 76)", "2ed5xwpn": "rgb(83, 135, 221)"}
    )
    run_id = (
        ops.project("stacey", "mendeleev")
        .filteredRuns("{}", "-createdAt")
        .limit(50)
        .summary()["table"]
        .table()
        .rows()
        .dropna()
        .concat()
        .createIndexCheckpointTag()[5]
        .run()
        .id()
    )
    run_color = colors_node[run_id]
    assert weave.use(run_color) == "rgb(83, 135, 221)"


def test_domain_gql_fragments(fake_wandb):
    fake_wandb.fake_api.add_mock(
        lambda q, ndx: {
            "project_518fa79465d8ffaeb91015dce87e092f": {
                **fwb.project_payload,
                "artifactType_46d22fef09db004187bb8da4b5e98c58": {
                    **fwb.defaultArtifactType_payload,
                    "artifactCollections_c1233b7003317090ab5e2a75db4ad965": {
                        "edges": [
                            {
                                "node": {
                                    **fwb.artifactSequence_payload,
                                    "artifacts_c1233b7003317090ab5e2a75db4ad965": {
                                        "edges": [
                                            {
                                                "node": {
                                                    **fwb.artifactVersion_payload,
                                                    "createdBy": {
                                                        "__typename": "Run",
                                                        **fwb.run_payload,
                                                        "displayName": "amber-glade-100",
                                                    },
                                                }
                                            }
                                        ]
                                    },
                                }
                            }
                        ]
                    },
                },
            }
        }
    )
    node = (
        ops.project("stacey", "mendeleev")
        .artifactType("test_results")
        .artifacts()[0]
        .versions()[0]
        .createdBy()
        .name()
    )
    assert weave.use(node) == "amber-glade-100"


def test_domain_gql_through_dicts_with_fn_nodes(fake_wandb, cache_mode_minimal):
    fake_wandb.fake_api.add_mock(lambda q, ndx: workspace_response())
    project_node = ops.project("stacey", "mendeleev")
    project_name_node = project_node.name()
    entity_node = project_node.entity()
    entity_name_node = entity_node.name()
    runs_node = project_node.runs()

    # Make a function that returns a dict of the name of the row
    var_node = graph.VarNode(wdt.RunType, "row")
    inner_fn = graph.ConstNode(
        types.Function({"row": wdt.RunType}, types.TypedDict({"name": types.String()})),
        ops.dict_(name=ops.run_ops.run_name(var_node)),
    )

    run_names = runs_node.map(inner_fn)

    dict_node = ops.dict_(
        project_name=project_name_node,
        entity_name=entity_name_node,
        run_names=run_names,
    )
    assert weave.use(dict_node) == {
        "project_name": "mendeleev",
        "entity_name": "stacey",
        "run_names": [
            {"name": "amber-glade-100"},
            {"name": "run2-display_name"},
            {"name": "run3-display_name"},
        ],
    }


def test_domain_gql_around_fn_nodes(fake_wandb):
    fake_wandb.fake_api.add_mock(lambda q, ndx: workspace_response())
    project_node = ops.project("stacey", "mendeleev")
    runs_node = project_node.runs()
    sorted_runs = runs_node.sort(lambda row: ops.make_list(a=row.createdAt()), ["asc"])
    names_node = sorted_runs.name()

    p = stitch.stitch([names_node])
    obj_recorder = p.get_result(runs_node)
    assert len(obj_recorder.calls) == 2


def test_lambda_gql_stitch(fake_wandb):
    fake_wandb.fake_api.add_mock(
        lambda q, ndx: {"project_518fa79465d8ffaeb91015dce87e092f": fwb.project_payload}
    )
    weave.use(
        ops.make_list(a="a", b="b", c="c").map(
            lambda x: x + ops.project("stacey", "mendeleev").name()
        )
    ) == None


def test_arrow_groupby_nested_tag_stripping(fake_wandb):
    fake_wandb.fake_api.add_mock(lambda q, ndx: file_path_response)
    groupby_node = (
        ops.project("stacey", "mendeleev")
        .artifactType("test_results")
        .artifacts()[0]
        .versions()[0]
        .file("test_results.table.json")
        .table()
        .rows()
        .createIndexCheckpointTag()
        .groupby(lambda row: ops.dict_(x=row["truth"]))[0]["truth"]
    )
    grouped = weave.use(groupby_node)
    assert len(grouped) == 50


def test_arrow_groupby_sort(fake_wandb):
    fake_wandb.fake_api.add_mock(lambda q, ndx: file_path_response)
    groupby_node = (
        ops.project("stacey", "mendeleev")
        .artifactType("test_results")
        .artifacts()[0]
        .versions()[0]
        .file("test_results.table.json")
        .table()
        .rows()
        .createIndexCheckpointTag()
        .groupby(lambda row: ops.dict_(x=row["truth"]))
        .sort(lambda row: ops.make_list(a=row.groupkey()["x"]), ["desc"])[0]
        .groupkey()["x"]
    )
    grouped = weave.use(groupby_node)
    assert grouped == "Reptilia"


def test_arrow_tag_stripping(fake_wandb):
    fake_wandb.fake_api.add_mock(lambda q, ndx: file_path_response)
    awl_node = (
        ops.project("stacey", "mendeleev")
        .artifactType("test_results")
        .artifacts()[0]
        .versions()[0]
        .file("test_results.table.json")
        .table()
        .rows()
        .createIndexCheckpointTag()
    )
    awl = weave.use(awl_node)
    tag_stripped = awl._arrow_data_asarray_no_tags()
    assert tag_stripped.type[0].name != "_tag"


def test_arrow_tag_serialization_can_handle_runs_in_concat(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock_filtered)
    rows_node = (
        ops.project("stacey", "mendeleev")
        .filteredRuns("{}", "-createdAt")
        .limit(50)
        .summary()["table"]
        .table()
        .rows()
    )

    const_list = ops.make_list(l=rows_node, r=rows_node)
    concatted_list = list_.List.concat(const_list)
    concatted = arrow.ops.concat(arr=concatted_list)

    # now get the run from the tags
    weave.use(ops.run_ops.run_tag_getter_op(concatted[0]))


@pytest.mark.skip(reason="This test requires communcation with wandb.ai")
def test_shawn_groupby_profiling_correctness():
    x = (
        ops.project("shawn", "dsviz-simple-tables")
        .artifact("simple_tables")
        .membershipForAlias("v5")
        .artifactVersion()
        .file("my-table.table.json")
        .table()
        .rows()
        .createIndexCheckpointTag()
        .groupby(lambda row: ops.dict_(x=row["x"]))[0]["c"]
    )

    cProfile.runctx(
        "weave.use(x)",
        globals(),
        locals(),
        filename="groupby-dsviz-simple-tables.pstat",
    )

    assert len(weave.use(x)) == 66311


def test_loading_artifact_browser_request_1(fake_wandb):
    # Leaf 1: Get's the current artifact collection
    fake_wandb.fake_api.add_mock(
        lambda q, ndx: {
            "project_518fa79465d8ffaeb91015dce87e092f": {
                **fwb.project_payload,  # type: ignore
                "artifactCollection_d651817074b6a8074e87e9dfd5767726": {
                    **fwb.artifactSequence_payload,  # type: ignore
                },
            }
        }
    )
    ac_node = ops.project("stacey", "mendeleev").artifact("test_res_1fwmcd3q")
    assert weave.use(ac_node) != None


def test_loading_artifact_browser_request_2(fake_wandb, cache_mode_minimal):
    ac_node = ops.project("stacey", "mendeleev").artifact("test_res_1fwmcd3q")
    # Leaf 2: Get collection details
    fake_wandb.fake_api.add_mock(
        lambda q, ndx: {
            "project_518fa79465d8ffaeb91015dce87e092f": {
                **fwb.project_payload,  # type: ignore
                "artifactCollection_d651817074b6a8074e87e9dfd5767726": {
                    **fwb.artifactSequence_payload,  # type: ignore
                    "__typename": "ArtifactSequence",
                    "project": fwb.project_payload,
                    "artifacts_c1233b7003317090ab5e2a75db4ad965": {
                        "edges": [{"node": {**fwb.artifactVersion_payload}}]
                    },
                    "artifactMembership_fe9cc269bee939ccb54ebba88c6087dd": {
                        **fwb.artifactMembership_payload
                    },
                    "artifactMemberships_first_1": {
                        "edges": [
                            {
                                "node": {
                                    **fwb.artifactMembership_payload,
                                    "aliases": [fwb.artifactAlias_payload],
                                }
                            }
                        ]
                    },
                    "artifactMemberships": {
                        "edges": [
                            {
                                "node": {
                                    **fwb.artifactMembership_payload,
                                    "aliases": [fwb.artifactAlias_payload],
                                }
                            }
                        ]
                    },
                },
            }
        }
    )
    ac_detail_node = dict_(
        **{
            "name": ac_node.name(),
            "isPortfolio": ac_node.isPortfolio(),
            "versionCount": ac_node.versions().count(),
            "latestVersionIndex": ac_node.lastMembership().versionIndex(),
            "projectName": ac_node.project().name(),
            "entityName": ac_node.project().entity().name(),
            "artifactTypeName": ac_node._get_op("type")().name(),
            "memberships": ac_node.memberships().map(
                lambda row: dict_(
                    **{
                        "versionIndex": row.versionIndex(),
                        "artifactVersion": row.artifactVersion(),
                        "aliases": row.aliases().alias(),
                    }
                )
            ),
        }
    )
    assert weave.use(ac_detail_node) != None

    # Leaf 3: Get the specific membership
    fake_wandb.fake_api.clear_mock_handlers()
    fake_wandb.fake_api.add_mock(
        lambda q, ndx: {
            "project_518fa79465d8ffaeb91015dce87e092f": {
                **fwb.project_payload,  # type: ignore
                "artifactCollection_d651817074b6a8074e87e9dfd5767726": {
                    **fwb.artifactSequence_payload,  # type: ignore
                    "artifactMembership_78e7a3fd51159b4fdfb0815be0b0f92c": {
                        **fwb.artifactMembership_payload
                    },
                },
            }
        }
    )
    mem_node = ac_node.membershipForAlias("v0")
    assert weave.use(mem_node) != None

    # Leaf 4: Get the specific membership details
    fake_wandb.fake_api.clear_mock_handlers()
    fake_wandb.fake_api.add_mock(
        lambda q, ndx: {
            "project_518fa79465d8ffaeb91015dce87e092f": {
                **fwb.project_payload,  # type: ignore
                "artifactCollection_d651817074b6a8074e87e9dfd5767726": {
                    **fwb.artifactSequence_payload,  # type: ignore
                    "artifactMembership_78e7a3fd51159b4fdfb0815be0b0f92c": {
                        **fwb.artifactMembership_payload
                    },
                },
            }
        }
    )
    mem_detail_node = dict_(
        **{
            "versionIndex": mem_node.versionIndex(),
            "artifactVersionId": mem_node.artifactVersion().id(),
        }
    )
    assert weave.use(mem_detail_node) != None


def test_loading_artifact_browser_request_3(fake_wandb, cache_mode_minimal):
    ac_node = ops.project("stacey", "mendeleev").artifact("test_res_1fwmcd3q")
    mem_node = ac_node.membershipForAlias("v0")
    # Leaf 5: Get the specific artifact version details
    fake_wandb.fake_api.add_mock(
        lambda q, ndx: {
            "project_518fa79465d8ffaeb91015dce87e092f": {
                **fwb.project_payload,  # type: ignore
                "artifactCollection_d651817074b6a8074e87e9dfd5767726": {
                    **fwb.artifactSequence_payload,  # type: ignore
                    "artifactMembership_78e7a3fd51159b4fdfb0815be0b0f92c": {
                        **fwb.artifactMembership_payload,  # type: ignore
                        "artifact": {
                            **fwb.artifactVersion_payload,  # type: ignore
                            "description": "",
                            "digest": "51560154fe3bae863d18335d39129732",
                            "commitHash": "303db33c9f9264768626",
                            "createdAt": "2021-07-10T19:27:32",
                            "usedBy": {
                                "edges": [
                                    {
                                        "node": {
                                            **fwb.run_payload,  # type: ignore
                                        }
                                    }
                                ]
                            },
                            "size": 90574246,
                            "artifactCollections": {
                                "edges": [
                                    {
                                        "node": {
                                            **fwb.artifactSequence_payload,  # type: ignore
                                            "__typename": "ArtifactSequence",
                                            "project": {
                                                **fwb.project_payload,  # type: ignore
                                            },
                                        }
                                    }
                                ]
                            },
                            "createdBy": {
                                "__typename": "Run",
                                **fwb.run_payload,  # type: ignore
                                "inputArtifacts_c1233b7003317090ab5e2a75db4ad965": {
                                    "edges": [
                                        {
                                            "node": {
                                                **fwb.artifactVersion_payload,  # type: ignore
                                                "artifactSequence": {
                                                    **fwb.artifactSequence_payload,  # type: ignore
                                                    "project": {
                                                        **fwb.project_payload,  # type: ignore
                                                    },
                                                },
                                                "artifactType": {
                                                    **fwb.defaultArtifactType_payload,  # type: ignore
                                                },
                                            },
                                        }
                                    ]
                                },
                            },
                        },
                        "artifactCollection": {
                            **fwb.artifactSequence_payload,  # type: ignore
                            "aliases": {
                                "edges": [
                                    {
                                        "node": {
                                            **fwb.artifactAlias_payload,  # type: ignore
                                        }
                                    }
                                ]
                            },
                            "project": {
                                **fwb.project_payload,  # type: ignore
                                "entity": {
                                    **fwb.entity_payload,  # type: ignore
                                    "artifactCollections_c96697489c051b1be46673088f743964": {
                                        "edges": [
                                            {
                                                "node": {
                                                    **fwb.artifactSequence_payload,  # type: ignore
                                                }
                                            }
                                        ]
                                    },
                                },
                            },
                            "__typename": "ArtifactSequence",
                            "artifactMemberships": {
                                "edges": [
                                    {
                                        "node": {
                                            **fwb.artifactMembership_payload,  # type: ignore
                                            "aliases": [
                                                {
                                                    **fwb.artifactAlias_payload,  # type: ignore
                                                },
                                            ],
                                        }
                                    }
                                ]
                            },
                        },
                    },
                },
            }
        }
    )
    av_node = mem_node.artifactVersion()
    mem_collection_node = mem_node.collection()
    av_details_node = dict_(
        **{
            "id": av_node.id(),
            "colId": mem_collection_node.id(),
            "colAliases": mem_collection_node.aliases().alias(),
            "colTypeName": mem_collection_node._get_op("type")().name(),
            "colEntityName": mem_collection_node.project().entity().name(),
            "colProjectName": mem_collection_node.project().name(),
            "colName": mem_collection_node.name(),
            "memVersionIndex": mem_node.versionIndex(),
            "description": av_node.description(),
            "digest": av_node.digest(),
            "createdAt": av_node.createdAt(),
            "usedByCount": av_node.usedBy().count(),
            "fileCount": av_node.files().count(),
            "size": av_node.size(),
            "isPortfolio": mem_collection_node.isPortfolio(),
            "memberships": mem_collection_node.memberships().map(
                lambda row: dict_(
                    **{
                        "versionIndex": row.versionIndex(),
                        "aliases": row.aliases().alias(),
                    }
                )
            ),
            "filteredCollections": av_node.artifactCollections()
            .filter(lambda row: row.isPortfolio())
            .map(
                lambda row: dict_(
                    **{
                        "name": row._get_op("name")(),
                        "typeName": row._get_op("type")().name(),
                        "entityName": row.project().entity().name(),
                        "projectName": row.project().name(),
                    }
                )
            ),
            "dependencies": av_node.createdBy()
            .usedArtifactVersions()
            .map(
                lambda row: dict_(
                    **{
                        "name": row._get_op("name")(),
                        "typeName": row.artifactType().name(),
                        "entityName": row.artifactSequence().project().entity().name(),
                        "projectName": row.artifactSequence().project().name(),
                    }
                )
            ),
            "hasPortfolios": mem_collection_node.project()
            .entity()
            .portfolios()
            .filter(
                lambda row: row._get_op("type")().name()
                == mem_collection_node._get_op("type")().name()
            )
            .count()
            >= 1,
        }
    )
    assert weave.use(av_details_node) != None

    # Leaf 6: Get Portfolios
    fake_wandb.fake_api.clear_mock_handlers()
    fake_wandb.fake_api.add_mock(
        lambda q, ndx: {
            "project_518fa79465d8ffaeb91015dce87e092f": {
                **fwb.project_payload,  # type: ignore
                "artifactCollection_d651817074b6a8074e87e9dfd5767726": {
                    **fwb.artifactSequence_payload,  # type: ignore
                    "artifactMembership_78e7a3fd51159b4fdfb0815be0b0f92c": {
                        **fwb.artifactMembership_payload,  # type: ignore
                        "artifact": {
                            **fwb.artifactVersion_payload,  # type: ignore
                            "artifactMemberships": {
                                "edges": [
                                    {
                                        "node": {
                                            **fwb.artifactMembership_payload,  # type: ignore
                                            "artifactCollection": {
                                                **fwb.artifactSequence_payload,  # type: ignore
                                                "__typename": "ArtifactSequence",
                                            },
                                        }
                                    }
                                ]
                            },
                        },
                    },
                },
            }
        }
    )
    portfolios_node = (
        av_node.memberships().filter(lambda row: row.collection().isPortfolio()).count()
    )
    assert weave.use(portfolios_node) != None

    # Leaf 7: Is Weave
    fake_wandb.fake_api.clear_mock_handlers()
    fake_wandb.fake_api.add_mock(
        lambda q, ndx: {
            "project_518fa79465d8ffaeb91015dce87e092f": {
                **fwb.project_payload,  # type: ignore
                "artifactCollection_d651817074b6a8074e87e9dfd5767726": {
                    **fwb.artifactSequence_payload,  # type: ignore
                    "artifactMembership_78e7a3fd51159b4fdfb0815be0b0f92c": {
                        **fwb.artifactMembership_payload,  # type: ignore
                        "artifact": {
                            **fwb.artifactVersion_payload,  # type: ignore
                        },
                    },
                },
            }
        }
    )
    is_weave_node = av_node.isWeaveObject()
    assert weave.use(is_weave_node) != None


example_history = [
    '{"_step":0,"loss":0.9416526556015015,"_runtime":247,"accuracy":0.7200000286102295,"epoch":0,"val_accuracy":0.7530242204666138,"_timestamp":1625961050,"val_loss":0.7652576565742493}',
    '{"_step":1,"predictions_10K":{"nrows":1000,"path":"media/table/predictions_10K_1_661a2d7d82a08afe8583.table.json","sha256":"661a2d7d82a08afe858320cc73ff3ac3ff3e5915a958c02e765963409d4716e9","size":147790,"_type":"table-file","artifact_path":"wandb-artifact://41727469666163743a3134323538383933/predictions_10K.table.json","ncols":14},"_runtime":294,"_timestamp":1625961097}',
    '{"_step":2,"loss":0.6638407707214355,"_runtime":545,"accuracy":0.7916250228881836,"epoch":1,"val_accuracy":0.7641128897666931,"_timestamp":1625961348,"val_loss":0.7551828622817993}',
    '{"_step":3,"predictions_10K":{"size":146810,"_type":"table-file","artifact_path":"wandb-artifact://41727469666163743a3134323539373631/predictions_10K.table.json","ncols":14,"nrows":1000,"path":"media/table/predictions_10K_3_3247b2550f0007cbaa5f.table.json","sha256":"3247b2550f0007cbaa5f3ad934d5d16bcbf966bcfc2770eb1e98e03b78b671fe"},"_runtime":591,"_timestamp":1625961394}',
    '{"_step":4,"loss":0.6015360951423645,"_runtime":824,"accuracy":0.8076249957084656,"epoch":2,"val_accuracy":0.7560483813285828,"_timestamp":1625961627,"val_loss":0.7426469326019287}',
    '{"_step":5,"predictions_10K":{"nrows":1000,"path":"media/table/predictions_10K_5_fa5fdbb0e0e48b873473.table.json","sha256":"fa5fdbb0e0e48b8734733a6a8bd2f460dfc3de35c1fedba5f0ac97eb9438ac17","size":146138,"_type":"table-file","artifact_path":"wandb-artifact://41727469666163743a3134323630353439/predictions_10K.table.json","ncols":14},"_runtime":870,"_timestamp":1625961673}',
    '{"_step":6,"loss":0.5667485594749451,"_runtime":1103,"accuracy":0.8165000081062317,"epoch":3,"val_accuracy":0.7752016186714172,"_timestamp":1625961906,"val_loss":0.7002165913581848}',
    '{"_step":7,"predictions_10K":{"ncols":14,"nrows":1000,"path":"media/table/predictions_10K_7_674d715bcfebaf8cc33e.table.json","sha256":"674d715bcfebaf8cc33ec920791498cae6321da393cc4bb8e44fd487eca4fedf","size":144151,"_type":"table-file","artifact_path":"wandb-artifact://41727469666163743a3134323631353532/predictions_10K.table.json"},"_runtime":1150,"_timestamp":1625961953}',
    '{"_step":8,"loss":0.5219206809997559,"_runtime":1383,"accuracy":0.8242499828338623,"epoch":4,"val_accuracy":0.78125,"_timestamp":1625962186,"val_loss":0.6415265202522278}',
    '{"_step":9,"predictions_10K":{"ncols":14,"nrows":1000,"path":"media/table/predictions_10K_9_fb0f1f25a0be1a907ec0.table.json","sha256":"fb0f1f25a0be1a907ec0c88efb482078736a40e49055e5251e42a368c12c9b2a","size":144982,"_type":"table-file","artifact_path":"wandb-artifact://41727469666163743a3134323632393738/predictions_10K.table.json"},"_runtime":1427,"_timestamp":1625962230}',
]
example_history_keys = {
    "sets": [],
    "keys": {
        "_step": {
            "typeCounts": [{"type": "number", "count": 10}],
            "monotonic": True,
            "previousValue": 9,
        },
        "_runtime": {
            "typeCounts": [{"type": "number", "count": 10}],
            "monotonic": True,
            "previousValue": 1427,
        },
        "_timestamp": {
            "typeCounts": [{"type": "number", "count": 10}],
            "monotonic": True,
            "previousValue": 1625962230,
        },
        "epoch": {
            "typeCounts": [{"type": "number", "count": 5}],
            "monotonic": True,
            "previousValue": 4,
        },
        "predictions_10K": {
            "typeCounts": [{"type": "table-file", "count": 5}],
            "monotonic": True,
            "previousValue": -1.7976931348623157e308,
        },
    },
    "lastStep": 9,
}

history_keys_with_media = {
    "sets": [],
    "keys": {
        "metric2": {
            "typeCounts": [{"type": "number", "count": 1001}],
            "previousValue": 1.6397776348651e06,
        },
        "_step": {
            "typeCounts": [{"type": "number", "count": 1001}],
            "monotonic": True,
            "previousValue": 1000,
        },
        "_runtime": {
            "typeCounts": [{"type": "number", "count": 1001}],
            "monotonic": True,
            "previousValue": 791.536298751831,
        },
        "_timestamp": {
            "typeCounts": [{"type": "number", "count": 1001}],
            "monotonic": True,
            "previousValue": 1.68322130708908e09,
        },
        "img": {
            "typeCounts": [
                {
                    "type": "image-file",
                    "count": 1001,
                    "keys": {
                        "boxes": [
                            {
                                "type": "map",
                                "keys": {
                                    "predictions": [
                                        {
                                            "type": "boxes2D",
                                            "keys": {
                                                "size": [{"type": "number"}],
                                                "_type": [{"type": "string"}],
                                                "sha256": [{"type": "string"}],
                                                "path": [{"type": "string"}],
                                            },
                                        }
                                    ]
                                },
                            }
                        ]
                    },
                    "nestedTypes": ["default"],
                }
            ],
            "monotonic": True,
            "previousValue": -1.7976931348623157e308,
        },
    },
    "lastStep": 1000,
}


def run_history_mocker(q, ndx):
    query_str = q["gql"].loc.source.body
    if "artifact" in query_str:
        return artifact_version_sdk_response

    node = {
        **fwb.run_payload,  # type: ignore
    }

    if "sampledParquetHistory" in query_str:
        node["sampledParquetHistory"] = {
            "parquetUrls": ["https://api.wandb.test/test.parquet"],
            "liveData": [
                json.loads(s)
                for s in [
                    '{"_step":8,"loss":0.5219206809997559,"_runtime":1383,"accuracy":0.8242499828338623,"epoch":4,"val_accuracy":0.78125,"_timestamp":1625962186,"val_loss":0.6415265202522278}',
                    '{"_step":9,"predictions_10K":{"ncols":14,"nrows":1000,"path":"media/table/predictions_10K_9_fb0f1f25a0be1a907ec0.table.json","sha256":"fb0f1f25a0be1a907ec0c88efb482078736a40e49055e5251e42a368c12c9b2a","size":144982,"_type":"table-file","artifact_path":"wandb-artifact://41727469666163743a3134323632393738/predictions_10K.table.json"},"_runtime":1427,"_timestamp":1625962230}',
                ]
            ],
        }

    if "historyKeys" in query_str:
        node["historyKeys"] = example_history_keys

    return {
        "project_518fa79465d8ffaeb91015dce87e092f": {
            **fwb.project_payload,  # type: ignore
            "runs_c1233b7003317090ab5e2a75db4ad965": {"edges": [{"node": node}]},
        }
    }


def run_history_mocker_with_pq_media(q, ndx):
    query_str = q["gql"].loc.source.body
    if "artifact" in query_str:
        return artifact_version_sdk_response

    node = {
        **fwb.run_payload,  # type: ignore
    }

    if "sampledParquetHistory" in query_str:
        node["sampledParquetHistory"] = {
            "parquetUrls": ["https://api.wandb.test/test_media.parquet"],
            "liveData": [
                {
                    "metric2": 1639777.6348651,
                    "_timestamp": 1683221307.0890799,
                    "_step": 1000.0,
                    "_runtime": 791.536298751831,
                    "img": {
                        "size": 3151148.0,
                        "height": 1024.0,
                        "width": 1024.0,
                        "_type": "image-file",
                        "format": "png",
                        "boxes": {
                            "predictions": {
                                "path": "media/metadata/boxes2D/img_999_7d232aa991dca7a6e3c5.boxes2D.json",
                                "size": 551.0,
                                "_type": "boxes2D",
                                "sha256": "7d232aa991dca7a6e3c5cd18acc3d3c7da1a865eb6b135f6c9f7af966b9fdfdb",
                            }
                        },
                        "path": "media/images/img_999_051320f86e77a481d041.png",
                        "sha256": "051320f86e77a481d041cd317fd2ab66925a2db8218c5e0c6fb61d740f24e8b5",
                    },
                }
            ],
        }

    if "historyKeys" in query_str:
        node["historyKeys"] = history_keys_with_media

    return {
        "project_518fa79465d8ffaeb91015dce87e092f": {
            **fwb.project_payload,  # type: ignore
            "runs_c1233b7003317090ab5e2a75db4ad965": {"edges": [{"node": node}]},
        }
    }


def test_run_history(fake_wandb):
    fake_wandb.fake_api.add_mock(run_history_mocker)
    node = ops.project("stacey", "mendeleev").runs()[0].history()
    assert isinstance(node.type, TaggedValueType)
    assert types.List(
        types.TypedDict(
            {
                "_step": types.Number(),
                # we no longer send back system metrics
                # "system/gpu.0.powerWatts": types.optional(types.Number()),
                "epoch": types.optional(types.Number()),
                "predictions_10K": types.optional(
                    artifact_fs.FilesystemArtifactFileType(
                        types.Const(types.String(), "json"), table.TableType()
                    ),
                ),
            }
        )
    ).assign_type(node.type.value)

    # this should just return dummy history
    assert weave.use(node) == [{"_step": i} for i in range(10)]

    # now we'll fetch a just one of the columns
    new_node = node["epoch"]
    assert types.List(
        types.optional(types.Number()),
    ).assign_type(new_node.type.value)
    assert weave.use(new_node) == [0, None, 1, None, 2, None, 3, None, 4, None]

    # now we'll create a graph that fetches two columns and execute it
    epoch_node = node["epoch"]
    pred_node = node["predictions_10K"]
    assert types.List(
        types.optional(
            artifact_fs.FilesystemArtifactFileType(
                types.Const(types.String(), "json"), table.TableType()
            ),
        )
    ).assign_type(pred_node.type.value)

    # simulates what a table call would do, selecting multiple columns
    result = weave.use([epoch_node, pred_node])
    assert len(result) == 2

    # check the types, result lengths
    assert all(len(r) == 10 for r in result)
    assert all(r == None or isinstance(r, (int, float)) for r in result[0])
    assert all(
        r == None or isinstance(r, artifact_fs.FilesystemArtifactFile)
        for r in result[1]
    )


def test_run_history_2(fake_wandb):
    fake_wandb.fake_api.add_mock(run_history_mocker)
    node = ops.project("stacey", "mendeleev").runs()[0].history2()
    assert isinstance(node.type, TaggedValueType)
    assert ArrowWeaveListType(
        types.TypedDict(
            {
                "_step": types.Number(),
                # we no longer send back system metrics
                # "system/gpu.0.powerWatts": types.optional(types.Number()),
                "epoch": types.optional(types.Number()),
                "predictions_10K": types.optional(
                    artifact_fs.FilesystemArtifactFileType(
                        types.Const(types.String(), "json"), table.TableType()
                    ),
                ),
            }
        )
    ).assign_type(node.type.value)

    # this should just return dummy history
    assert weave.use(node).to_pylist_raw() == [{"_step": i} for i in range(10)]

    # now we'll fetch a just one of the columns
    new_node = node["epoch"]
    assert ArrowWeaveListType(
        types.optional(types.Number()),
    ).assign_type(new_node.type.value)
    assert weave.use(new_node).to_pylist_raw() == [
        0,
        None,
        1,
        None,
        2,
        None,
        3,
        None,
        4,
        None,
    ]

    # now we'll create a graph that fetches all three columns and execute it
    epoch_node = node["epoch"]
    pred_node = node["_runtime"]
    table_node = node["predictions_10K"]
    assert types.List(types.optional(types.Number())).assign_type(pred_node.type.value)

    # simulates what a table call would do, selecting multiple columns
    result = weave.use([epoch_node, pred_node, table_node])
    assert len(result) == 3

    # check the types, result lengths
    assert all(len(r) == 10 for r in result)
    assert all(r == None or isinstance(r, (int, float)) for r in result[0])
    assert all(r == None or isinstance(r, (int, float)) for r in result[1])
    assert all(
        r == None or isinstance(r, artifact_fs.FilesystemArtifactFile)
        for r in result[2]
    )


def test_run_history2_media_types(fake_wandb, cache_mode_minimal):
    fake_wandb.fake_api.add_mock(run_history_mocker_with_pq_media)

    # this is actually from the launch-test/prodmon project, but mocked here as stacey/mendeleev
    node = ops.project("stacey", "mendeleev").runs()[0].history2()
    image_node = node["img"]
    metric2_node = node["metric2"]

    # simulates what a table call would do, selecting multiple columns
    result = weave.use([image_node, metric2_node])

    assert types.List(types.optional(wbmedia.ImageArtifactFileRefType())).assign_type(
        image_node.type
    )
    assert types.List(types.optional(types.Number())).assign_type(metric2_node.type)
    assert len(result) == 2
    assert len(result[0]) == 1001

    # check that parquet data is converted properly
    assert isinstance(result[0][0], wbmedia.ImageArtifactFileRef)

    # check that live data is converted properly
    last_img = result[0][len(result[0]) - 1]
    assert isinstance(last_img, wbmedia.ImageArtifactFileRef)
    assert last_img.boxes == {}
    assert last_img.masks == {}
    assert last_img.classes is None
    assert last_img.path == "media/images/img_999_051320f86e77a481d041.png"
    assert last_img.format == "png"
    assert last_img.width == 1024
    assert last_img.height == 1024
    assert (
        last_img.sha256
        == "051320f86e77a481d041cd317fd2ab66925a2db8218c5e0c6fb61d740f24e8b5"
    )


def run_history_as_of_mocker(q, ndx):
    return {
        "project_518fa79465d8ffaeb91015dce87e092f": {
            **fwb.project_payload,  # type: ignore
            "runs_c1233b7003317090ab5e2a75db4ad965": {
                "edges": [
                    {
                        "node": {
                            **fwb.run_payload,  # type: ignore
                            "history_c81e728d9d4c2f636f067f89cc14862c": [
                                example_history[2]
                            ],
                        }
                    }
                ]
            },
        }
    }


def test_run_history_as_of(fake_wandb):
    fake_wandb.fake_api.add_mock(run_history_as_of_mocker)
    node = ops.project("stacey", "mendeleev").runs()[0].historyAsOf(2)
    assert isinstance(node.type, TaggedValueType)
    assert types.TypedDict(
        {
            "_step": types.Number(),
            "_timestamp": types.Number(),
            "_runtime": types.Number(),
            "accuracy": types.Number(),
            "epoch": types.Number(),
            "loss": types.Number(),
            "val_accuracy": types.Number(),
            "val_loss": types.Number(),
        }
    ).assign_type(node.type.value)

    keys = ["_step", "_timestamp", "_runtime", "predictions_10K"]
    nodes = [node[key] for key in keys]

    assert weave.use(nodes) == [
        json.loads(example_history[2]).get(key, None) for key in keys
    ]


def test_artifact_membership_link(fake_wandb):
    fake_wandb.fake_api.add_mock(lambda q, ndx: artifact_browser_response)
    node = amo.artifact_membership_link(
        ops.project("stacey", "mendeleev")
        .artifact("test_res_1fwmcd3q")
        .membershipForAlias("v0")
    )

    assert weave.use(node) == wdt.Link(
        name="test_res_1fwmcd3q:v0",
        url="/stacey/mendeleev/artifacts/test_results/test_res_1fwmcd3q/v0",
    )


def test_run_colors(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock_filtered)
    colors_node = ops.dict_(
        **{"1ht5692d": "rgb(218, 76, 76)", "2ed5xwpn": "rgb(83, 135, 221)"}
    )
    node = (
        ops.project("stacey", "mendeleev")
        .filteredRuns("{}", "-createdAt")
        .limit(50)
        .summary()
        .pick("table")
        .table()
        .rows()
        .dropna()
        .concat()
        .map(
            lambda row: ops.dict_(
                color=colors_node.pick(row.run().id()), name=row.run().name()
            )
        )
        .index(0)
        .pick("color")
    )
    assert weave.use(node) == "rgb(83, 135, 221)"


def test_arrow_groupby_external_tag(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock_filtered)
    run_names = (
        ops.project("stacey", "mendeleev")
        .filteredRuns("{}", "-createdAt")
        .limit(50)
        .summary()
        .pick("table")
        .table()
        .rows()
        .dropna()
        .concat()
        .createIndexCheckpointTag()
        .groupby(lambda row: ops.dict_(x=row["truth"]))
        .sort(lambda row: ops.make_list(a=row.groupkey()["x"]), ["desc"])[0]
        .run()
        .name()[0]
    )
    run_names_res = weave.use(run_names)
    assert run_names_res == "amber-glade-100"


def test_join_all_tables(fake_wandb, cache_mode_minimal):
    fake_wandb.fake_api.add_mock(table_mock_filtered)
    joined_row = (
        ops.project("stacey", "mendeleev")
        .filteredRuns("{}", "-createdAt")
        .limit(50)
        .summary()
        .pick("table")
        .table()
        .rows()
        .dropna()
        .joinAll(lambda row: row["truth"], True)
        .createIndexCheckpointTag()[0]
    )
    run_names_res = weave.use(joined_row["score_Fungi"][0].run().name())
    assert run_names_res == "amber-glade-100"
    run_names_res = weave.use(joined_row["score_Fungi"][0].project().name())
    assert run_names_res == "mendeleev"


def test_arrow_unnest_shallow_tags(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock_filtered)
    x_val = (
        ops.project("stacey", "mendeleev")
        .filteredRuns("{}", "-createdAt")
        .limit(50)[0]
        .summary()
        .pick("table")
        .table()
        .rows()
        .map(lambda row: ops.dict_(x=row["truth"], y=row["guess"]))
        .unnest()[3]["x"]
    )
    run_name = x_val.run().name()


def test_arrow_unnest_deep_tags(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock_filtered)
    x_val = (
        ops.project("stacey", "mendeleev")
        .filteredRuns("{}", "-createdAt")
        .limit(50)
        .summary()
        .pick("table")
        .table()
        .rows()
        .dropna()
        .concat()
        .createIndexCheckpointTag()
        .map(lambda row: ops.dict_(x=row["truth"], y=row["guess"]))
        .unnest()[3]["x"]
    )
    run_name = x_val.run().name()
    assert weave.use(run_name) == "amber-glade-100"

    checkpoint_tag = x_val.indexCheckpoint()
    assert weave.use(checkpoint_tag) == 3

    assert weave.use(run_name) == "amber-glade-100"


def test_arrow_unnest_inner_tags(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock_filtered)
    run_name = (
        (
            ops.project("stacey", "mendeleev")
            .filteredRuns("{}", "-createdAt")
            .limit(50)
            .summary()
            .pick("table")
            .table()
            .rows()
        )
        .joinAll(lambda row: row["truth"], True)
        .map(lambda row: ops.dict_(x=row["truth"], y=row["guess"]))
        .unnest()[0]["x"]
        .run()
        .name()
    )
    assert weave.use(run_name) == "amber-glade-100"


def test_image_sha_from_table(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock2)
    img_node = (
        ops.project("stacey", "mendeleev")
        .runs()[0]
        .summary()["table"]
        .table()
        .rows()[0]["image"]
    )
    assert (
        weave.use(img_node.sha256)
        == "50a7232cb6785ffae981038af58de306192e5739a0b9d9903fa436f9f508e7d5"
    )


def test_image_in_summary(fake_wandb):
    fake_wandb.fake_api.add_mock(table_mock2)

    img_node = ops.project("stacey", "mendeleev").runs()[0].summary()["image"]

    assert (
        weave.use(img_node).sha256
        == "440fab0d6f537b4557a106fa7853453332650631ef580fd328c620bd8aa5a025"
    )


@pytest.mark.parametrize(
    "input_uri, expected_uri",
    [
        (
            "wandb-logged-artifact://long_server_id:latest/path",
            "wandb-artifact:///random-entity/random-project/run-2isjqtcr-Validation_table:303db33c9f9264768626/path",
        ),
        (
            "wandb-logged-artifact://long_server_id:v4/path",
            "wandb-artifact:///random-entity/random-project/run-2isjqtcr-Validation_table:303db33c9f9264768626/path",
        ),
        (
            "wandb-logged-artifact://1234567890/path",
            "wandb-artifact:///stacey/mendeleev/test_res_1fwmcd3q:303db33c9f9264768626/path",
        ),
        (
            "wandb-artifact:///input-entity/input-project/run-2isjqtcr-Validation_table/path",
            "wandb-artifact:///input-entity/input-project/run-2isjqtcr-Validation_table:303db33c9f9264768626/path",
        ),
    ],
)
def test_wb_artifact_uri_resolution(fake_wandb, input_uri, expected_uri):
    if input_uri == "wandb-logged-artifact://1234567890/path":
        fake_wandb.fake_api.add_mock(lambda q, index: artifact_version_sdk_response)
    else:
        fake_wandb.fake_api.add_mock(
            lambda q, index: {
                "artifactCollection": {
                    "id": "QXJ0aWZhY3RDb2xsZWN0aW9uOjUzNDYxMTc2",
                    "name": "run-2isjqtcr-Validation_table",
                    "state": "READY",
                    "project": {
                        "id": "UHJvamVjdDp2MTpNZXJnZWRfbGlndG5pbmc6Y29udGFjdC1lc3RpbWF0aW9u",
                        "name": "random-project",
                        "entity": {
                            "id": "RW50aXR5OjUwODA2Mg==",
                            "name": "random-entity",
                        },
                    },
                    "artifactMembership": {
                        "id": "QXJ0aWZhY3RDb2xsZWN0aW9uTWVtYmVyc2hpcDozNzAxNzE5NDE=",
                        "versionIndex": 4,
                        "commitHash": "303db33c9f9264768626",
                    },
                    "defaultArtifactType": {
                        "id": "QXJ0aWZhY3RUeXBlOjM1OTgxNA==",
                        "name": "run_table",
                    },
                }
            }
        )

    uri = uris.WeaveURI.parse(input_uri)
    ref = artifact_wandb.WandbArtifactRef.from_uri(uri=uri)
    assert str(ref) == expected_uri


def test_is_valid_version_string():
    assert not artifact_wandb.is_valid_version_index("latest")

    for v in ["v0", "v1", "v10", "v19"]:
        assert artifact_wandb.is_valid_version_index(v)

    for v in ["v0.0", "v1.0", "v10.0", "v19.0"]:
        assert not artifact_wandb.is_valid_version_index(v)

    for v in ["v01", "v0009"]:
        assert not artifact_wandb.is_valid_version_index(v)


def test_artifact_path_character_escaping():
    name = 12347187287418787843872388177814
    path = "table #3.table.json"
    result = wb_util.escape_artifact_path(
        f"wandb-client-artifact://{name}:latest/{path}"
    )
    uri = artifact_wandb.WeaveWBLoggedArtifactURI.parse(result)

    assert uri.path == path


def _do_test_gql_artifact_dir_path(node):
    dir_node = node.file("")
    dir_type_node = dir_node.pathReturnType("")
    dir_file_type_node = dir_node.pathReturnType("test_results.table.json")

    file_node = dir_node.path("test_results.table.json")
    file_type_node = file_node._get_op("type")()

    assert artifact_fs.FilesystemArtifactDirType().assign_type(weave.use(dir_type_node))
    assert artifact_fs.FilesystemArtifactFileType(
        weave.types.Const(weave.types.String(), "json"), wbObjectType=ops.TableType()
    ).assign_type(weave.use(dir_file_type_node))
    assert artifact_fs.FilesystemArtifactFileType(
        weave.types.Const(weave.types.String(), "json"), wbObjectType=ops.TableType()
    ).assign_type(weave.use(file_type_node))


def test_gql_artifact_dir_path(fake_wandb):
    fake_wandb.fake_api.add_mock(lambda q, ndx: artifact_browser_response)
    version_node = (
        ops.project("stacey", "mendeleev")
        .artifact("test_res_1fwmcd3q")
        .membershipForAlias("v0")
        .artifactVersion()
    )
    _do_test_gql_artifact_dir_path(version_node)


def test_filesystem_artifact_dir_path(fake_wandb):
    table = wandb.Table(data=[[1]], columns=["a"])
    art = wandb.Artifact("test_name", "test_type")
    art.add(table, "test_results")
    art_node = fake_wandb.mock_artifact_as_node(art)
    _do_test_gql_artifact_dir_path(art_node)


def test_filesystem_artifact_dir_dict(fake_wandb):
    table = wandb.Table(
        data=[[1, wandb.Image(np.zeros((32, 32)))]], columns=["a", "image"]
    )
    art = wandb.Artifact("test_name", "test_type")
    art.add(table, "test_results")
    art.add_reference("https://www.google.com", "google_link")
    art_node = fake_wandb.mock_artifact_as_node(art)
    assert weave.use(_as_w0_dict_(art_node.file(""))) == {
        "fullPath": "",
        "size": 954,
        "dirs": {
            "media": {
                "fullPath": "media/images/842c574e85966d3269f6.png",
                "size": 75,
                "dirs": {"images": 1},
                "files": {},
            }
        },
        "files": {
            "test_results.table.json": {
                "birthArtifactID": "TODO",
                "digest": "I/aUoc5M4r1UgTNwJUECMA==",
                "fullPath": "test_results.table.json",
                "size": 879,
                "type": "file",
                "url": "https://api.wandb.ai/artifacts/test_entity/23f694a1ce4ce2bd5481337025410230/test_results.table.json",
            },
            "google_link": {
                "birthArtifactID": "TODO",
                "digest": "https://www.google.com",
                "fullPath": "google_link",
                "size": 0,
                "type": "file",
                "ref": "https://www.google.com",
                "url": "https://www.google.com",
            },
        },
    }


def test_filesystem_artifact_by_id_dir_dict(fake_wandb):
    table = wandb.Table(
        data=[[1, wandb.Image(np.zeros((32, 32)))]], columns=["a", "image"]
    )
    art = wandb.Artifact("test_name", "test_type")
    art.add(table, "test_results")
    art.add_reference("https://www.google.com", "google_link")
    art_node = fake_wandb.mock_artifact_by_id_as_node(art)
    assert weave.use(_as_w0_dict_(art_node.file(""))) == {
        "fullPath": "",
        "size": 954,
        "dirs": {
            "media": {
                "fullPath": "media/images/842c574e85966d3269f6.png",
                "size": 75,
                "dirs": {"images": 1},
                "files": {},
            }
        },
        "files": {
            "test_results.table.json": {
                "birthArtifactID": "TODO",
                "digest": "I/aUoc5M4r1UgTNwJUECMA==",
                "fullPath": "test_results.table.json",
                "size": 879,
                "type": "file",
                "url": "https://api.wandb.ai/artifacts/_/23f694a1ce4ce2bd5481337025410230/test_results.table.json",
            },
            "google_link": {
                "birthArtifactID": "TODO",
                "digest": "https://www.google.com",
                "fullPath": "google_link",
                "size": 0,
                "type": "file",
                "ref": "https://www.google.com",
                "url": "https://www.google.com",
            },
        },
    }


def test_non_const_input_node(fake_wandb, cache_mode_minimal):
    fake_wandb.fake_api.add_mock(table_mock1)
    data_obj = weave.save({"project_name": "mendeleev"})
    cell_node = (
        # projectName arg is an OutputNode. Compile should execute
        # the branch to convert it to a ConstNode, so the gql compile
        # pass can see the value.
        ops.project("stacey", data_obj["project_name"])
        .runs()
        .limit(1)
        .summary()["table"]
        .table()
        .rows()
        .dropna()
        .concat()
        .createIndexCheckpointTag()[5]["score_Amphibia"]
    )
    assert weave.use(cell_node.indexCheckpoint()) == 5
    assert weave.use(cell_node.run().name()) == "amber-glade-100"
    assert weave.use(cell_node.project().name()) == "mendeleev"



import typing

import datasets
import huggingface_hub

import weave


class HFDatasetInfoTypedDict(typing.TypedDict):
    id: str
    lastModified: str
    tags: list[str]
    description: str
    # citation: str
    # sha: str
    # downloads: int


@weave.op(render_info={"type": "function"})
def hf_datasets() -> list[HFDatasetInfoTypedDict]:
    api = huggingface_hub.HfApi()
    return [
        {
            "id": dataset_info.id,
            "lastModified": dataset_info.lastModified,
            "tags": dataset_info.tags or [],
            "description": dataset_info.description,
            # "downloads": dataset_info.downloads,
        }
        for dataset_info in api.list_datasets()
    ]


# TODO: we really need finer grained numeric types!
def hf_feature_type_to_type(type_):
    if isinstance(type_, dict):
        prop_types = {}
        for key, type_ in type_.items():
            prop_types[key] = hf_feature_type_to_type(type_)
        return weave.types.TypedDict(prop_types)
    elif isinstance(type_, datasets.features.features.Sequence):
        return weave.types.List(hf_feature_type_to_type(type_.feature))
    elif isinstance(type_, datasets.features.features.Image):
        return weave.legacy.weave.ops.PILImageType()
    elif isinstance(type_, datasets.features.features.ClassLabel):
        # TODO: this should be a classes type!!!!!
        return weave.types.Int()
    else:
        dtype = type_.dtype
        if dtype == "int8":
            return weave.types.Int()
        elif dtype == "int16":
            return weave.types.Int()
        elif dtype == "int32":
            return weave.types.Int()
        elif dtype == "int64":
            return weave.types.Int()
        elif dtype == "float32":
            return weave.types.Float()
        elif dtype == "string":
            return weave.types.String()
    raise weave.errors.WeaveTypeError("unhandled hf type: %s" % type_)


@weave.op(render_info={"type": "function"}, hidden=True)
def dataset_refine_output_type(name: str) -> weave.types.Type:
    ds = datasets.load_dataset(name, split="train", streaming=True)
    prop_types = {}
    for key, type_ in ds.features.items():
        prop_types[key] = hf_feature_type_to_type(type_)
    return weave.types.List(weave.types.TypedDict(prop_types))


@weave.op(
    render_info={"type": "function"},
    output_type=weave.types.List(weave.types.TypedDict({})),
    refine_output_type=dataset_refine_output_type,
)
def dataset(name: str):
    ds = datasets.load_dataset(name, split="train", streaming=True)

    rows = []
    for _, row in zip(range(100), iter(ds)):
        rows.append(row)
    return rows



# from ..ecosystem import keras
import json
import typing
from dataclasses import dataclass

import pytest

from weave.legacy.weave import api as weave
from weave.legacy.weave import context_state

_loading_builtins_token = context_state.set_loading_built_ins()


@dataclass(frozen=True)
class CustomType(weave.types.Type):
    type_var_1: weave.types.Type = weave.types.Any()
    type_var_2: weave.types.Type = weave.types.Any()

    @weave.op(output_type=lambda input_types: input_types["self"].type_var_1)
    def get_1(self):
        return self.type_var_1


@weave.op(
    input_type={"obj": CustomType()},
    output_type=lambda input_types: input_types["obj"][input_types["x"].val],
)
def custom_op(obj, x: str):
    if x == "type_var_1":
        return obj.type_var_1
    elif x == "type_var_2":
        return obj.type_var_2
    else:
        return None


@weave.op(name="int_concat")
def int_concat(a: int, b: int) -> str:
    return str(a) + str(b)


@weave.op(output_type=lambda input_types: input_types["obj"])
def identity(obj: typing.Any):
    return obj


@weave.op(
    output_type=lambda input_types: weave.types.List.make(
        {"object_type": input_types["obj"]}
    )
)
def wrap(obj: typing.Any):
    return [obj]


context_state.clear_loading_built_ins(_loading_builtins_token)

test_data = json.load(open("./legacy/tests/test_op_def_data.json"))


def test_op_def_to_dict():
    assert int_concat.to_dict() == test_data["op_defs"]["int_concat"]["def"]


@pytest.mark.skip("callable output type serialization disabled")
def test_identity():
    assert identity.to_dict() == test_data["op_defs"]["op-identity"]["def"]


@pytest.mark.skip("callable output type serialization disabled")
def test_make_list():
    assert wrap.to_dict() == test_data["op_defs"]["op-wrap"]["def"]


@pytest.mark.skip("callable output type serialization disabled")
def test_custom_datatype():
    assert CustomType.get_1.to_dict() == test_data["op_defs"]["op-get_1"]["def"]
    assert custom_op.to_dict() == test_data["op_defs"]["op-custom_op"]["def"]


@pytest.mark.skip("callable output type serialization disabled")
def test_keras_model():
    assert keras.call_string.to_dict() == test_data["op_defs"]["op-call_string"]["def"]



# This file should include tests for all the fixups we do to match
# the frontend js code.
# The related sites should be tagged with "# NOTE: js_compat" in the
#     Weave python code.
# TODO: Fix all the non-standard behaviors in the weave js code so we
#     can get rid of this file.
# Note: This file is not yet complete, there are existing fixups in the
#     weave Python code that I haven't documented here.

from weave.legacy.weave import partial_object, weavejs_fixes
from weave.legacy.weave import weave_types as types
from weave.legacy.weave.ops_domain import wb_domain_types


def test_const_serialization():
    f_type = types.FileType(types.Const(types.String(), "png"))
    f_type_dict = f_type.to_dict()
    assert f_type_dict == {"extension": "png", "type": "FileBase"}
    f_type2 = types.TypeRegistry.type_from_dict(f_type_dict)
    assert isinstance(f_type2, types.FileType)
    assert isinstance(f_type2.extension, types.Const)
    assert f_type2.extension.val == "png"


def test_partialobject_type_stripping():
    instance = wb_domain_types.Run({"a": types.String()})
    type = types.TypeRegistry.type_of(instance)
    assert isinstance(type, partial_object.PartialObjectType)
    serialized = type.to_dict()
    assert weavejs_fixes.remove_partialobject_from_types(serialized) == "run"


def test_nested_partialobject_type_stripping():
    input = {
        "type": "tagged",
        "tag": {
            "type": "typedDict",
            "propertyTypes": {
                "project": {
                    "type": "PartialObject",
                    "keys": {
                        "id": "string",
                        "name": "string",
                        "runs_b7f2f1441a601b3477d356429a493d05": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "edges": {
                                    "type": "list",
                                    "objectType": {
                                        "type": "typedDict",
                                        "propertyTypes": {
                                            "node": {
                                                "type": "typedDict",
                                                "propertyTypes": {
                                                    "id": "string",
                                                    "name": "string",
                                                    "summaryMetrics": "string",
                                                },
                                            }
                                        },
                                    },
                                }
                            },
                        },
                    },
                    "keyless_weave_type_class": "project",
                }
            },
        },
        "value": {
            "type": "list",
            "objectType": {
                "type": "union",
                "members": [
                    {
                        "type": "tagged",
                        "tag": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "run": {
                                    "type": "PartialObject",
                                    "keys": {
                                        "id": "string",
                                        "name": "string",
                                        "summaryMetrics": "string",
                                    },
                                    "keyless_weave_type_class": "run",
                                }
                            },
                        },
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "epoch": "int",
                                "_timestamp": "float",
                                "samples_wf_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "val_mean_absolute_error": "float",
                                "graph": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "sha256": "string",
                                        "path": "string",
                                        "size": "int",
                                        "_type": "string",
                                    },
                                },
                                "best_epoch": "int",
                                "samples_wf_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "final_val_loss": "float",
                                "_step": "int",
                                "val_mae": "float",
                                "_runtime": "float",
                                "test_loss": "float",
                                "samples_wf_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "mean_absolute_error": "float",
                                "loss": "float",
                                "GFLOPs": "int",
                                "test_mae": "float",
                                "val_loss": "float",
                                "best_val_loss": "float",
                            },
                        },
                    },
                    {
                        "type": "tagged",
                        "tag": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "run": {
                                    "type": "PartialObject",
                                    "keys": {
                                        "id": "string",
                                        "name": "string",
                                        "summaryMetrics": "string",
                                    },
                                    "keyless_weave_type_class": "run",
                                }
                            },
                        },
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "_step": "int",
                                "_runtime": "float",
                                "samples_wf_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "_timestamp": "float",
                                "samples_wf_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                            },
                        },
                    },
                    {
                        "type": "tagged",
                        "tag": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "run": {
                                    "type": "PartialObject",
                                    "keys": {
                                        "id": "string",
                                        "name": "string",
                                        "summaryMetrics": "string",
                                    },
                                    "keyless_weave_type_class": "run",
                                }
                            },
                        },
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "samples_wf_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "best_val_loss": "float",
                                "final_val_loss": "float",
                                "graph": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "sha256": "string",
                                        "path": "string",
                                        "size": "int",
                                        "_type": "string",
                                    },
                                },
                                "best_epoch": "int",
                                "_timestamp": "float",
                                "samples_wf_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "val_mean_absolute_error": "float",
                                "epoch": "int",
                                "test_mae": "float",
                                "val_loss": "float",
                                "loss": "float",
                                "_runtime": "float",
                                "mean_absolute_error": "float",
                                "samples_wf_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "_step": "int",
                                "val_mae": "float",
                                "test_loss": "float",
                            },
                        },
                    },
                    {
                        "type": "tagged",
                        "tag": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "run": {
                                    "type": "PartialObject",
                                    "keys": {
                                        "id": "string",
                                        "name": "string",
                                        "summaryMetrics": "string",
                                    },
                                    "keyless_weave_type_class": "run",
                                }
                            },
                        },
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "samples_wf_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "_runtime": "float",
                                "test_mae": "float",
                                "val_mae": "float",
                                "test_loss": "float",
                                "_timestamp": "float",
                                "samples_wf_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "final_val_loss": "float",
                                "_step": "int",
                            },
                        },
                    },
                    {
                        "type": "tagged",
                        "tag": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "run": {
                                    "type": "PartialObject",
                                    "keys": {
                                        "id": "string",
                                        "name": "string",
                                        "summaryMetrics": "string",
                                    },
                                    "keyless_weave_type_class": "run",
                                }
                            },
                        },
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "samples_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "_timestamp": "float",
                                "_step": "int",
                                "_runtime": "float",
                                "samples_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                            },
                        },
                    },
                    {
                        "type": "tagged",
                        "tag": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "run": {
                                    "type": "PartialObject",
                                    "keys": {
                                        "id": "string",
                                        "name": "string",
                                        "summaryMetrics": "string",
                                    },
                                    "keyless_weave_type_class": "run",
                                }
                            },
                        },
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "loss": "float",
                                "GFLOPs": "int",
                                "_runtime": "float",
                                "samples_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "best_epoch": "int",
                                "val_mean_absolute_error": "float",
                                "test_loss": "float",
                                "_step": "int",
                                "epoch": "int",
                                "graph": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "path": "string",
                                        "size": "int",
                                        "_type": "string",
                                        "sha256": "string",
                                    },
                                },
                                "val_loss": "float",
                                "samples_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "_timestamp": "float",
                                "best_val_loss": "float",
                                "final_val_loss": "float",
                                "mean_absolute_error": "float",
                                "val_mae": "float",
                                "test_mae": "float",
                            },
                        },
                    },
                    {
                        "type": "tagged",
                        "tag": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "run": {
                                    "type": "PartialObject",
                                    "keys": {
                                        "id": "string",
                                        "name": "string",
                                        "summaryMetrics": "string",
                                    },
                                    "keyless_weave_type_class": "run",
                                }
                            },
                        },
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "_step": "int",
                                "graph": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "path": "string",
                                        "size": "int",
                                        "_type": "string",
                                        "sha256": "string",
                                    },
                                },
                                "samples_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "val_mean_absolute_error": "float",
                                "samples_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "final_val_loss": "float",
                                "epoch": "int",
                                "_runtime": "float",
                                "test_mae": "float",
                                "samples_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "_timestamp": "float",
                                "best_epoch": "int",
                                "best_val_loss": "float",
                                "loss": "float",
                                "val_mae": "float",
                                "val_loss": "float",
                                "samples_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "test_loss": "float",
                                "mean_absolute_error": "float",
                            },
                        },
                    },
                    {
                        "type": "tagged",
                        "tag": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "run": {
                                    "type": "PartialObject",
                                    "keys": {
                                        "id": "string",
                                        "name": "string",
                                        "summaryMetrics": "string",
                                    },
                                    "keyless_weave_type_class": "run",
                                }
                            },
                        },
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "val_mae": "float",
                                "samples_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "test_loss": "float",
                                "_timestamp": "float",
                                "_step": "int",
                                "_runtime": "float",
                                "test_mae": "float",
                                "samples_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "final_val_loss": "float",
                            },
                        },
                    },
                ],
            },
        },
    }

    actual = weavejs_fixes.remove_partialobject_from_types(input)

    expected = {
        "type": "tagged",
        "tag": {"type": "typedDict", "propertyTypes": {"project": "project"}},
        "value": {
            "type": "list",
            "objectType": {
                "type": "union",
                "members": [
                    {
                        "type": "tagged",
                        "tag": {"type": "typedDict", "propertyTypes": {"run": "run"}},
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "epoch": "int",
                                "_timestamp": "float",
                                "samples_wf_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "val_mean_absolute_error": "float",
                                "graph": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "sha256": "string",
                                        "path": "string",
                                        "size": "int",
                                        "_type": "string",
                                    },
                                },
                                "best_epoch": "int",
                                "samples_wf_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "final_val_loss": "float",
                                "_step": "int",
                                "val_mae": "float",
                                "_runtime": "float",
                                "test_loss": "float",
                                "samples_wf_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "mean_absolute_error": "float",
                                "loss": "float",
                                "GFLOPs": "int",
                                "test_mae": "float",
                                "val_loss": "float",
                                "best_val_loss": "float",
                            },
                        },
                    },
                    {
                        "type": "tagged",
                        "tag": {"type": "typedDict", "propertyTypes": {"run": "run"}},
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "_step": "int",
                                "_runtime": "float",
                                "samples_wf_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "_timestamp": "float",
                                "samples_wf_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                            },
                        },
                    },
                    {
                        "type": "tagged",
                        "tag": {"type": "typedDict", "propertyTypes": {"run": "run"}},
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "samples_wf_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "best_val_loss": "float",
                                "final_val_loss": "float",
                                "graph": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "sha256": "string",
                                        "path": "string",
                                        "size": "int",
                                        "_type": "string",
                                    },
                                },
                                "best_epoch": "int",
                                "_timestamp": "float",
                                "samples_wf_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "val_mean_absolute_error": "float",
                                "epoch": "int",
                                "test_mae": "float",
                                "val_loss": "float",
                                "loss": "float",
                                "_runtime": "float",
                                "mean_absolute_error": "float",
                                "samples_wf_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "_step": "int",
                                "val_mae": "float",
                                "test_loss": "float",
                            },
                        },
                    },
                    {
                        "type": "tagged",
                        "tag": {"type": "typedDict", "propertyTypes": {"run": "run"}},
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "samples_wf_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "_runtime": "float",
                                "test_mae": "float",
                                "val_mae": "float",
                                "test_loss": "float",
                                "_timestamp": "float",
                                "samples_wf_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_wf_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "final_val_loss": "float",
                                "_step": "int",
                            },
                        },
                    },
                    {
                        "type": "tagged",
                        "tag": {"type": "typedDict", "propertyTypes": {"run": "run"}},
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "samples_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "_timestamp": "float",
                                "_step": "int",
                                "_runtime": "float",
                                "samples_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                            },
                        },
                    },
                    {
                        "type": "tagged",
                        "tag": {"type": "typedDict", "propertyTypes": {"run": "run"}},
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "loss": "float",
                                "GFLOPs": "int",
                                "_runtime": "float",
                                "samples_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "best_epoch": "int",
                                "val_mean_absolute_error": "float",
                                "test_loss": "float",
                                "_step": "int",
                                "epoch": "int",
                                "graph": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "path": "string",
                                        "size": "int",
                                        "_type": "string",
                                        "sha256": "string",
                                    },
                                },
                                "val_loss": "float",
                                "samples_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "_timestamp": "float",
                                "best_val_loss": "float",
                                "final_val_loss": "float",
                                "mean_absolute_error": "float",
                                "val_mae": "float",
                                "test_mae": "float",
                            },
                        },
                    },
                    {
                        "type": "tagged",
                        "tag": {"type": "typedDict", "propertyTypes": {"run": "run"}},
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "_step": "int",
                                "graph": {
                                    "type": "typedDict",
                                    "propertyTypes": {
                                        "path": "string",
                                        "size": "int",
                                        "_type": "string",
                                        "sha256": "string",
                                    },
                                },
                                "samples_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "val_mean_absolute_error": "float",
                                "samples_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "final_val_loss": "float",
                                "epoch": "int",
                                "_runtime": "float",
                                "test_mae": "float",
                                "samples_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "_timestamp": "float",
                                "best_epoch": "int",
                                "best_val_loss": "float",
                                "loss": "float",
                                "val_mae": "float",
                                "val_loss": "float",
                                "samples_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "test_loss": "float",
                                "mean_absolute_error": "float",
                            },
                        },
                    },
                    {
                        "type": "tagged",
                        "tag": {"type": "typedDict", "propertyTypes": {"run": "run"}},
                        "value": {
                            "type": "typedDict",
                            "propertyTypes": {
                                "val_mae": "float",
                                "samples_0": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_1": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_3": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "samples_4": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "test_loss": "float",
                                "_timestamp": "float",
                                "_step": "int",
                                "_runtime": "float",
                                "test_mae": "float",
                                "samples_2": {
                                    "type": "file",
                                    "_base_type": {"type": "FileBase"},
                                    "extension": "json",
                                    "wbObjectType": {
                                        "type": "table",
                                        "_base_type": {"type": "Object"},
                                        "_is_object": True,
                                        "_rows": {
                                            "type": "ArrowWeaveList",
                                            "_base_type": {"type": "list"},
                                            "objectType": {
                                                "type": "typedDict",
                                                "propertyTypes": {},
                                            },
                                        },
                                    },
                                },
                                "final_val_loss": "float",
                            },
                        },
                    },
                ],
            },
        },
    }

    assert actual == expected



import weave
import dataclasses


class TweetType(weave.types.ObjectType):
    def property_types(self):
        return {
            "_created_at": weave.types.String(),
            "_id": weave.types.Int(),
            "_text": weave.types.String(),
            "_truncated": weave.types.Boolean(),
            "_possibly_sensitive": weave.types.Boolean(),
        }


@weave.weave_class(weave_type=TweetType)
@dataclasses.dataclass
class Tweet:
    _id: int
    _created_at: str
    _text: str
    _truncated: bool
    _possibly_sensitive: bool

    @weave.op()
    def id(self) -> int:
        return self._id

    @weave.op()
    def created_at(self) -> str:
        return self._created_at

    @weave.op()
    def text(self) -> str:
        return self._text

    @weave.op()
    def truncated(self) -> bool:
        return self._truncated

    @weave.op()
    def possibly_sensitive(self) -> bool:
        return self._possibly_sensitive


TweetType.instance_classes = Tweet



# Test for weave.show
# The outputs of weave.show (the generated panel urls and their arguments)
# need to match what javascript expects.

from weave.legacy.weave import ops

from ...legacy.weave.show import _show_params
from . import test_helpers


def test_show_simple_call(cereal_csv):
    csv = ops.local_path(cereal_csv).readcsv()
    actual = _show_params(csv)["weave_node"].to_json()
    print("test_show_simple_call.actual", actual)
    assert actual == {
        "nodeType": "output",
        "type": "any",
        "fromOp": {
            "name": "get",
            "inputs": {
                "uri": {
                    "nodeType": "const",
                    "type": "string",
                    "val": "local-artifact:///dashboard-table:latest/obj",
                }
            },
        },
    }
